@@ -1,31 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Quick 3D.
-**
-** $QT_BEGIN_LICENSE:GPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 or (at your option) any later version
-** approved by the KDE Free Qt Foundation. The licenses are as published by
-** the Free Software Foundation and appearing in the file LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2019 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include "qquick3dmodel_p.h"
 #include "qquick3dobject_p.h"
@@ -130,8 +104,6 @@ QQuick3DModel::~QQuick3DModel()
 {
     disconnect(m_skeletonConnection);
     disconnect(m_geometryConnection);
-    for (const auto &connection : qAsConst(m_connections))
-        disconnect(connection);
 
     auto matList = materials();
     qmlClearMaterials(&matList);
@@ -271,7 +243,10 @@ void QQuick3DModel::markAllDirty()
     \qmlproperty bool Model::castsShadows
 
     When this property is \c true, the geometry of this model is used when
-    rendering to the shadow maps.
+    rendering to the shadow maps, and is also generating shadows in baked
+    lighting.
+
+    The default value is \c true.
 */
 
 bool QQuick3DModel::castsShadows() const
@@ -282,8 +257,14 @@ bool QQuick3DModel::castsShadows() const
 /*!
     \qmlproperty bool Model::receivesShadows
 
-    When this property is set to \c true, the model's materials take shadow contribution from
-    shadow casting lights into account.
+    When this property is set to \c true, the model's materials take shadow
+    contribution from shadow casting lights into account.
+
+    \note When lightmapping is enabled for this model, fully baked lights with
+    Light::bakeMode set to Light.BakeModeAll will always generate (baked)
+    shadows on the model, regardless of the value of this property.
+
+    The default value is \c true.
 */
 
 bool QQuick3DModel::receivesShadows() const
@@ -334,6 +315,26 @@ QQuick3DSkeleton *QQuick3DModel::skeleton() const
     return m_skeleton;
 }
 
+/*!
+    \qmlproperty Skin Model::skin
+
+    Contains the skeleton for the model. The Skeleton is used for
+    \l {Vertex Skinning}{skinning}.
+
+    \note Meshes of the model must have both joints and weights attributes.
+    \note If this property is set, skinning animation is enabled. This means
+    the \l {Model} will be transformed based on \l {Skin::joints} and the
+    Model's global transformation will be ignored.
+    \note If a model has both a skeleton and a skin, then the skin will be used.
+
+    \sa {Model::skeleton} {Qt Quick 3D - Simple Skinning Example}
+*/
+QQuick3DSkin *QQuick3DModel::skin() const
+{
+    return m_skin;
+}
+
+
 /*!
     \qmlproperty List<matrix4x4> Model::inverseBindPoses
 
@@ -380,7 +381,6 @@ float QQuick3DModel::depthBias() const
 }
 
 /*!
-    \preliminary
     \qmlproperty bool Model::receivesReflections
 
     When this property is set to \c true, the model's materials take reflections contribution from
@@ -393,6 +393,120 @@ bool QQuick3DModel::receivesReflections() const
     return m_receivesReflections;
 }
 
+/*!
+    \qmlproperty bool Model::castsReflections
+    \since 6.4
+
+    When this property is set to \c true, the model is rendered by reflection probes and can be
+    seen in the reflections.
+*/
+bool QQuick3DModel::castsReflections() const
+{
+    return m_castsReflections;
+}
+
+/*!
+    \qmlproperty bool Model::usedInBakedLighting
+
+    When this property is set to \c true, the model contributes to baked
+    lighting, such as lightmaps, for example in form of casting shadows or
+    indirect light. This setting is independent of controlling lightmap
+    generation for the model, use \l bakedLightmap for that.
+
+    The default value is \c false.
+
+    \note The default value is false, because designers and developers must
+    always evaluate on a per-model basis if the object is suitable to take part
+    in baked lighting.
+
+    \warning Models with dynamically changing properties, for example, animated
+    position, rotation, or other properties, are not suitable for participating
+    in baked lighting.
+
+    For more information on how to bake lightmaps, see the \l Lightmapper
+    documentation.
+
+    This property is relevant only when baking lightmaps. It has no effect
+    afterwards, when using the generated lightmaps during rendering.
+
+    \sa Light::bakeMode, bakedLightmap, Lightmapper, {Lightmaps and Global Illumination}
+ */
+
+bool QQuick3DModel::isUsedInBakedLighting() const
+{
+    return m_usedInBakedLighting;
+}
+
+/*!
+    \qmlproperty int Model::lightmapBaseResolution
+
+    Defines the approximate size of the lightmap for this model. The default
+    value is 1024, indicating 1024x1024 as the base size. The actual size of
+    the lightmap texture is likely to be different, often bigger, depending on
+    the mesh.
+
+    For simpler, smaller meshes, or when it is known that using a bigger
+    lightmap is unnecessary, the value can be set to something smaller, for
+    example, 512 or 256.
+
+    The minimum value is 128.
+
+    This setting applies both to persistently stored and for intermediate,
+    partial lightmaps. When baking lightmaps, all models that have \l
+    usedInBakedLighting enabled are part of the path-traced scene. Thus all of
+    them need to have lightmap UV unwrapping performed and the rasterization
+    steps necessary to compute direct lighting which then can be taken into
+    account for indirect light bounces in the scene. However, for models that
+    just contribute to, but do not store a lightmap the default value is often
+    sufficient. Fine-tuning is more relevant for models that store and then use
+    the generated lightmaps.
+
+    This property is relevant only when baking lightmaps. It has no effect
+    afterwards, when using the generated lightmaps during rendering.
+
+    Models that have lightmap UV data pre-generated during asset import time
+    (e.g. via the balsam tool) will ignore this property because the lightmap
+    UV unwrapping and the lightmap size hint evaluation have already been done,
+    and will not be performed again during lightmap baking.
+ */
+int QQuick3DModel::lightmapBaseResolution() const
+{
+    return m_lightmapBaseResolution;
+}
+
+/*!
+    \qmlproperty BakedLightmap Model::bakedLightmap
+
+    When this property is set to a valid, enabled BakedLightmap object, the
+    model will get a lightmap generated when baking lighting, the lightmap is
+    then stored persistently. When rendering, the model will load and use the
+    associated lightmap. The default value is null.
+
+    \note When the intention is to generate a persistently stored lightmap for
+    a Model, both bakedLightmap and \l usedInBakedLighting must be set on it,
+    in order to indicate that the Model not only participates in the
+    lightmapped scene, but also wants a full lightmap to be baked and stored.
+
+    For more information on how to bake lightmaps, see the \l Lightmapper
+    documentation.
+
+    This property is relevant both when baking and when using lightmaps. A
+    consistent state between the baking run and the subsequent runs that use
+    the generated data is essential. For example, changing the lightmap key
+    will make it impossible to load the previously generated data. An exception
+    is \l enabled, which can be used to dynamically toggle the usage of
+    lightmaps (outside of the baking run), but be aware that the rendered
+    results will depend on the Lights' \l{Light::bakeMode}{bakeMode} settings
+    in the scene.
+
+    \sa usedInBakedLighting, Lightmapper
+ */
+
+QQuick3DBakedLightmap *QQuick3DModel::bakedLightmap() const
+{
+    return m_bakedLightmap;
+}
+
 void QQuick3DModel::setSource(const QUrl &source)
 {
     if (m_source == source)
@@ -441,9 +555,7 @@ void QQuick3DModel::setGeometry(QQuick3DGeometry *geometry)
         return;
 
     // Make sure to disconnect if the geometry gets deleted out from under us
-    QQuick3DObjectPrivate::updatePropertyListener(geometry, m_geometry, QQuick3DObjectPrivate::get(this)->sceneManager, QByteArrayLiteral("geometry"), m_connections, [this](QQuick3DObject *n) {
-        setGeometry(qobject_cast<QQuick3DGeometry *>(n));
-    });
+    QQuick3DObjectPrivate::attachWatcher(this, &QQuick3DModel::setGeometry, geometry, m_geometry);
 
     if (m_geometry)
         QObject::disconnect(m_geometryConnection);
@@ -465,9 +577,7 @@ void QQuick3DModel::setSkeleton(QQuick3DSkeleton *skeleton)
         return;
 
     // Make sure to disconnect if the skeleton gets deleted out from under us
-    QQuick3DObjectPrivate::updatePropertyListener(skeleton, m_skeleton, QQuick3DObjectPrivate::get(this)->sceneManager, QByteArrayLiteral("skeleton"), m_connections, [this](QQuick3DObject *n) {
-        setSkeleton(qobject_cast<QQuick3DSkeleton *>(n));
-    });
+    QQuick3DObjectPrivate::attachWatcher(this, &QQuick3DModel::setSkeleton, skeleton, m_skeleton);
 
     if (m_skeleton)
         QObject::disconnect(m_skeletonConnection);
@@ -484,6 +594,19 @@ void QQuick3DModel::setSkeleton(QQuick3DSkeleton *skeleton)
     markDirty(SkeletonDirty);
 }
 
+void QQuick3DModel::setSkin(QQuick3DSkin *skin)
+{
+    if (skin == m_skin)
+        return;
+
+    // Make sure to disconnect if the skin gets deleted out from under us
+    QQuick3DObjectPrivate::attachWatcher(this, &QQuick3DModel::setSkin, skin, m_skin);
+
+    m_skin = skin;
+    emit skinChanged();
+    markDirty(SkinDirty);
+}
+
 void QQuick3DModel::setInverseBindPoses(const QList<QMatrix4x4> &poses)
 {
     if (m_inverseBindPoses == poses)
@@ -510,9 +633,7 @@ void QQuick3DModel::setInstancing(QQuick3DInstancing *instancing)
         return;
 
     // Make sure to disconnect if the instance table gets deleted out from under us
-    QQuick3DObjectPrivate::updatePropertyListener(instancing, m_instancing, QQuick3DObjectPrivate::get(this)->sceneManager, QByteArrayLiteral("instancing"), m_connections, [this](QQuick3DObject *n) {
-        setInstancing(qobject_cast<QQuick3DInstancing *>(n));
-    });
+    QQuick3DObjectPrivate::attachWatcher(this, &QQuick3DModel::setInstancing, instancing, m_instancing);
     if (m_instancing)
         QObject::disconnect(m_instancingConnection);
     m_instancing = instancing;
@@ -554,6 +675,60 @@ void QQuick3DModel::setReceivesReflections(bool receivesReflections)
     markDirty(ReflectionDirty);
 }
 
+void QQuick3DModel::setCastsReflections(bool castsReflections)
+{
+    if (m_castsReflections == castsReflections)
+        return;
+    m_castsReflections = castsReflections;
+    emit castsReflectionsChanged();
+    markDirty(ReflectionDirty);
+}
+
+void QQuick3DModel::setUsedInBakedLighting(bool enable)
+{
+    if (m_usedInBakedLighting == enable)
+        return;
+
+    m_usedInBakedLighting = enable;
+    emit usedInBakedLightingChanged();
+    markDirty(PropertyDirty);
+}
+
+void QQuick3DModel::setLightmapBaseResolution(int resolution)
+{
+    resolution = qMax(128, resolution);
+    if (m_lightmapBaseResolution == resolution)
+        return;
+
+    m_lightmapBaseResolution = resolution;
+    emit lightmapBaseResolutionChanged();
+    markDirty(PropertyDirty);
+}
+
+void QQuick3DModel::setBakedLightmap(QQuick3DBakedLightmap *bakedLightmap)
+{
+    if (m_bakedLightmap == bakedLightmap)
+        return;
+
+    if (m_bakedLightmap)
+        m_bakedLightmap->disconnect(m_bakedLightmapSignalConnection);
+
+    m_bakedLightmap = bakedLightmap;
+
+    m_bakedLightmapSignalConnection = QObject::connect(m_bakedLightmap, &QQuick3DBakedLightmap::changed, this,
+                                                       [this] { markDirty(PropertyDirty); });
+
+    QObject::connect(m_bakedLightmap, &QObject::destroyed, this,
+                     [this]
+    {
+        m_bakedLightmap = nullptr;
+        markDirty(PropertyDirty);
+    });
+
+    emit bakedLightmapChanged();
+    markDirty(PropertyDirty);
+}
+
 void QQuick3DModel::itemChange(ItemChange change, const ItemChangeData &value)
 {
     if (change == QQuick3DObject::ItemSceneChange)
@@ -574,7 +749,7 @@ QSSGRenderGraphObject *QQuick3DModel::updateSpatialNode(QSSGRenderGraphObject *n
     if (m_dirtyAttributes & SourceDirty)
         modelNode->meshPath = QSSGRenderPath(translateMeshSource(m_source, this));
     if (m_dirtyAttributes & PickingDirty)
-        modelNode->flags.setFlag(QSSGRenderModel::Flag::LocallyPickable, m_pickable);
+        modelNode->setState(QSSGRenderModel::LocalState::Pickable, m_pickable);
 
     if (m_dirtyAttributes & ShadowsDirty) {
         modelNode->castsShadows = m_castsShadows;
@@ -613,7 +788,7 @@ QSSGRenderGraphObject *QQuick3DModel::updateSpatialNode(QSSGRenderGraphObject *n
             const int numMorphTarget = m_morphTargets.size();
             if (modelNode->morphTargets.isEmpty()) {
                 // Easy mode, just add each morphTarget
-                for (const auto morphTarget : qAsConst(m_morphTargets)) {
+                for (const auto morphTarget : std::as_const(m_morphTargets)) {
                     QSSGRenderGraphObject *graphObject = QQuick3DObjectPrivate::get(morphTarget)->spatialNode;
                     if (graphObject)
                         modelNode->morphTargets.append(graphObject);
@@ -664,16 +839,39 @@ QSSGRenderGraphObject *QQuick3DModel::updateSpatialNode(QSSGRenderGraphObject *n
             modelNode->skeleton = nullptr;
     }
 
+    if (m_dirtyAttributes & SkinDirty) {
+        if (m_skin)
+            modelNode->skin = static_cast<QSSGRenderSkin *>(QQuick3DObjectPrivate::get(m_skin)->spatialNode);
+        else
+            modelNode->skin = nullptr;
+    }
+
     if (m_dirtyAttributes & PoseDirty) {
         modelNode->inverseBindPoses = m_inverseBindPoses.toVector();
         modelNode->skinningDirty = true;
     }
 
-    if (m_dirtyAttributes & PropertyDirty)
+    if (m_dirtyAttributes & PropertyDirty) {
         modelNode->m_depthBias = m_depthBias;
+        modelNode->usedInBakedLighting = m_usedInBakedLighting;
+        modelNode->lightmapBaseResolution = uint(m_lightmapBaseResolution);
+        if (m_bakedLightmap && m_bakedLightmap->isEnabled()) {
+            modelNode->lightmapKey = m_bakedLightmap->key();
+            const QString srcPrefix = m_bakedLightmap->loadPrefix();
+            const QString srcPath = srcPrefix.isEmpty() ? QStringLiteral(".") : srcPrefix;
+            const QQmlContext *context = qmlContext(m_bakedLightmap);
+            const QUrl resolvedUrl = context ? context->resolvedUrl(srcPath) : srcPath;
+            modelNode->lightmapLoadPath = QQmlFile::urlToLocalFileOrQrc(resolvedUrl);
+        } else {
+            modelNode->lightmapKey.clear();
+            modelNode->lightmapLoadPath.clear();
+        }
+    }
 
-    if (m_dirtyAttributes & ReflectionDirty)
+    if (m_dirtyAttributes & ReflectionDirty) {
         modelNode->receivesReflections = m_receivesReflections;
+        modelNode->castsReflections = m_castsReflections;
+    }
 
     m_dirtyAttributes = dirtyAttribute;
 
@@ -693,6 +891,7 @@ void QQuick3DModel::updateSceneManager(QQuick3DSceneManager *sceneManager)
     if (sceneManager) {
         sceneManager->dirtyBoundingBoxList.append(this);
         QQuick3DObjectPrivate::refSceneManager(m_skeleton, *sceneManager);
+        QQuick3DObjectPrivate::refSceneManager(m_skin, *sceneManager);
         QQuick3DObjectPrivate::refSceneManager(m_geometry, *sceneManager);
         QQuick3DObjectPrivate::refSceneManager(m_instancing, *sceneManager);
         for (Material &mat : m_materials) {
@@ -705,6 +904,7 @@ void QQuick3DModel::updateSceneManager(QQuick3DSceneManager *sceneManager)
         }
     } else {
         QQuick3DObjectPrivate::derefSceneManager(m_skeleton);
+        QQuick3DObjectPrivate::derefSceneManager(m_skin);
         QQuick3DObjectPrivate::derefSceneManager(m_geometry);
         QQuick3DObjectPrivate::derefSceneManager(m_instancing);
         for (Material &mat : m_materials) {
@@ -719,7 +919,7 @@ void QQuick3DModel::updateSceneManager(QQuick3DSceneManager *sceneManager)
 void QQuick3DModel::onMaterialDestroyed(QObject *object)
 {
     bool found = false;
-    for (int i = 0; i < m_materials.count(); ++i) {
+    for (int i = 0; i < m_materials.size(); ++i) {
         if (m_materials[i].material == object) {
             m_materials.removeAt(i--);
             found = true;
@@ -771,7 +971,7 @@ QQuick3DMaterial *QQuick3DModel::qmlMaterialAt(QQmlListProperty<QQuick3DMaterial
 qsizetype QQuick3DModel::qmlMaterialsCount(QQmlListProperty<QQuick3DMaterial> *list)
 {
     QQuick3DModel *self = static_cast<QQuick3DModel *>(list->object);
-    return self->m_materials.count();
+    return self->m_materials.size();
 }
 
 void QQuick3DModel::qmlClearMaterials(QQmlListProperty<QQuick3DMaterial> *list)
@@ -845,13 +1045,13 @@ QQuick3DMorphTarget *QQuick3DModel::qmlMorphTargetAt(QQmlListProperty<QQuick3DMo
 qsizetype QQuick3DModel::qmlMorphTargetsCount(QQmlListProperty<QQuick3DMorphTarget> *list)
 {
     QQuick3DModel *self = static_cast<QQuick3DModel *>(list->object);
-    return self->m_morphTargets.count();
+    return self->m_morphTargets.size();
 }
 
 void QQuick3DModel::qmlClearMorphTargets(QQmlListProperty<QQuick3DMorphTarget> *list)
 {
     QQuick3DModel *self = static_cast<QQuick3DModel *>(list->object);
-    for (const auto &morph : qAsConst(self->m_morphTargets)) {
+    for (const auto &morph : std::as_const(self->m_morphTargets)) {
         if (morph->parentItem() == nullptr)
             QQuick3DObjectPrivate::get(morph)->derefSceneManager();
         morph->disconnect(self, SLOT(onMorphTargetDestroyed(QObject*)));
