@@ -1,31 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Quick 3D.
-**
-** $QT_BEGIN_LICENSE:GPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 or (at your option) any later version
-** approved by the KDE Free Qt Foundation. The licenses are as published by
-** the Free Software Foundation and appearing in the file LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2019 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #ifndef QSSGOBJECT_P_H
 #define QSSGOBJECT_P_H
@@ -65,19 +39,76 @@ class Q_QUICK3D_PRIVATE_EXPORT QQuick3DObjectPrivate : public QObjectPrivate
 public:
     using Type = QSSGRenderGraphObject::Type;
 
-    using ConnectionMap = QHash<QByteArray, QMetaObject::Connection>;
+    struct ConnectionKey
+    {
+        using Handle = void (QQuick3DObject::*)(QObject *);
+        QObject *context = nullptr;
+        Handle unusable = nullptr;
+        friend bool operator==(const ConnectionKey &a, const ConnectionKey &b) noexcept { return (a.context == b.context) && (a.unusable == b.unusable); }
+    };
+    using ConnectionMap = QHash<ConnectionKey, QMetaObject::Connection>;
+
+    template<typename SceneContext, typename CallContext, typename Setter>
+    static void attachWatcherPriv(SceneContext *sceneContext, CallContext *callContext, Setter setter, QQuick3DObject *newO, QObject *oldO)
+    {
+        static_assert(std::is_base_of_v<QQuick3DObject, SceneContext>, "The scene context must be a QQuick3DObject");
+        static_assert(std::is_member_function_pointer_v<Setter>, "The assumption is that the setter is a member function!");
+        static_assert(sizeof(ConnectionKey::Handle) >= sizeof(Setter), "The handle needs to be able to store the value of the setter");
+        auto sceneManager = QQuick3DObjectPrivate::get(sceneContext)->sceneManager;
+        auto &connectionMap = QQuick3DObjectPrivate::get(sceneContext)->connectionMap;
+        union
+        {
+            Setter s;
+            ConnectionKey::Handle h;
+        }; s = setter;
+        ConnectionKey key{static_cast<QObject *>(callContext), h};
+        // disconnect previous destruction listener
+        if (oldO) {
+            // NOTE: If the old object is inside the QObject's dtor (e.g., QObject::destroyed) we can't
+            // call deref (and there's no point anymore either).
+            if (auto old3dO = qobject_cast<QQuick3DObject *>(oldO))
+                QQuick3DObjectPrivate::derefSceneManager(old3dO);
+
+            auto it = connectionMap.constFind(key);
+            if (it != connectionMap.cend()) {
+                QObject::disconnect(*it);
+                connectionMap.erase(it);
+            }
+        }
+
+        // Watch new object
+        if (newO) {
+            if (sceneManager)
+                QQuick3DObjectPrivate::refSceneManager(newO, *sceneManager);
+            auto connection = QObject::connect(newO, &QObject::destroyed, callContext, [callContext, setter](){ (callContext->*setter)(nullptr); });
+            connectionMap.insert(key, connection);
+        }
+    }
+
+    /*!
+      Attach a object-destroyed-watcher to an object that's not owned.
+      There are few checks here just to keep it simple
+      (The compiler should still fail with a varying degree of helpful messages when used incorrectly).
+
+      \a sceneContext - ususally the same as the callContext and only different if the calledContext is a non-QQuick3DObject class
+      (as is the case for QQuick3DShaderUtilsTextureInput)!
+      \a callContext - The object watching another object
+      \a setter - The function/slot that is called for the object (context).
+      \a newO - The new object being watched
+      \b oldO - The previous object that should no longer be watched.
+
+      Note: The \a setter is a function that takes one argument with a discardable return value.
+    */
+    template<typename Context, typename Setter, typename Object3D>
+    static void attachWatcher(Context *context, Setter setter, Object3D *newO, Object3D *oldO)
+    {
+        attachWatcherPriv(context, context, setter, newO, oldO);
+    }
 
     static QQuick3DObjectPrivate *get(QQuick3DObject *item) { return item->d_func(); }
     static const QQuick3DObjectPrivate *get(const QQuick3DObject *item) { return item->d_func(); }
     static QSSGRenderGraphObject *updateSpatialNode(QQuick3DObject *o, QSSGRenderGraphObject *n) { return o->updateSpatialNode(n); }
 
-    static void updatePropertyListener(QQuick3DObject *newO,
-                                       QQuick3DObject *oldO,
-                                       QQuick3DSceneManager *sceneManager,
-                                       const QByteArray &propertyKey,
-                                       ConnectionMap &connections,
-                                       const std::function<void(QQuick3DObject *o)> &callFn);
-
     explicit QQuick3DObjectPrivate(Type t);
     ~QQuick3DObjectPrivate() override;
     void init(QQuick3DObject *parent);
@@ -250,11 +281,19 @@ public:
     bool componentComplete = true;
     bool preSyncNeeded = false;
     bool culled;
+    bool sharedResource = false;
     QQuick3DItem2D *contentItem2d = nullptr;
+    ConnectionMap connectionMap;
 };
 
 Q_DECLARE_OPERATORS_FOR_FLAGS(QQuick3DObjectPrivate::ChangeTypes)
 Q_DECLARE_TYPEINFO(QQuick3DObjectPrivate::ChangeListener, Q_PRIMITIVE_TYPE);
+Q_DECLARE_TYPEINFO(QQuick3DObjectPrivate::ConnectionKey, Q_PRIMITIVE_TYPE);
+
+inline size_t qHash(const QQuick3DObjectPrivate::ConnectionKey &con, size_t seed = 0) noexcept
+{
+    return qHashBits(&con, sizeof(QQuick3DObjectPrivate::ConnectionKey), seed);
+}
 
 QT_END_NAMESPACE
 
