@@ -1,31 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2019 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of Qt Quick 3D.
-**
-** $QT_BEGIN_LICENSE:GPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 or (at your option) any later version
-** approved by the KDE Free Qt Foundation. The licenses are as published by
-** the Free Software Foundation and appearing in the file LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2019 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include "qquick3dviewport_p.h"
 #include "qquick3dsceneenvironment_p.h"
@@ -41,6 +15,7 @@
 #include "qquick3ddefaultmaterial_p.h"
 #include "qquick3dprincipledmaterial_p.h"
 #include "qquick3dcustommaterial_p.h"
+#include "qquick3dspecularglossymaterial_p.h"
 #include <QtQuick3DRuntimeRender/private/qssgrenderlayer_p.h>
 
 #include <qsgtextureprovider.h>
@@ -52,10 +27,15 @@
 
 #include <QtQml>
 
+#include <QtGui/private/qeventpoint_p.h>
+
+#include <QtCore/private/qnumeric_p.h>
+
 QT_BEGIN_NAMESPACE
 
 Q_LOGGING_CATEGORY(lcEv, "qt.quick3d.event")
 Q_LOGGING_CATEGORY(lcPick, "qt.quick3d.pick")
+Q_LOGGING_CATEGORY(lcHover, "qt.quick3d.hover")
 
 struct ViewportTransformHelper : public QQuickDeliveryAgent::Transform
 {
@@ -72,21 +52,32 @@ struct ViewportTransformHelper : public QQuickDeliveryAgent::Transform
         owners.append(da);
     }
 
-    /*!
+    /*
         Transforms viewport coordinates to 2D scene coordinates.
         Returns the point in targetItem corresponding to \a viewportPoint,
         assuming that targetItem is mapped onto sceneParentNode.
+        If it's no longer a "hit" on sceneParentNode, returns the last-good point.
     */
     QPointF map(const QPointF &viewportPoint) override {
-        auto pickResult = renderer->syncPickOne(viewportPoint * dpr, sceneParentNode);
-        auto ret = pickResult.m_localUVCoords.toPointF();
-        if (!uvCoordsArePixels)
-            ret = QPointF(targetItem->x() + ret.x() * targetItem->width(),
-                          targetItem->y() - ret.y() * targetItem->height() + targetItem->height());
-
-        qCDebug(lcEv) << viewportPoint << "->" << ret << "@" << pickResult.m_scenePosition
-                      << "UV" << pickResult.m_localUVCoords << "dist" << qSqrt(pickResult.m_cameraDistanceSq);
-        return ret;
+        QSSGOption<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(viewportPoint * dpr);
+        if (rayResult.hasValue()) {
+            auto pickResult = renderer->syncPickOne(rayResult.getValue(), sceneParentNode);
+            auto ret = pickResult.m_localUVCoords.toPointF();
+            if (!uvCoordsArePixels) {
+                ret = QPointF(targetItem->x() + ret.x() * targetItem->width(),
+                              targetItem->y() - ret.y() * targetItem->height() + targetItem->height());
+            }
+            const bool outOfModel = pickResult.m_localUVCoords.isNull();
+            qCDebug(lcEv) << viewportPoint << "->" << (outOfModel ? "OOM" : "") << ret << "@" << pickResult.m_scenePosition
+                          << "UV" << pickResult.m_localUVCoords << "dist" << qSqrt(pickResult.m_distanceSq);
+            if (outOfModel) {
+                return lastGoodMapping;
+            } else {
+                lastGoodMapping = ret;
+                return ret;
+            }
+        }
+        return QPointF();
     }
 
     QQuick3DSceneRenderer *renderer = nullptr;
@@ -94,6 +85,7 @@ struct ViewportTransformHelper : public QQuickDeliveryAgent::Transform
     QQuickItem* targetItem = nullptr;
     qreal dpr = 1;
     bool uvCoordsArePixels = false; // if false, they are in the range 0..1
+    QPointF lastGoodMapping;
 
     static QList<QPointer<QQuickDeliveryAgent>> owners;
 };
@@ -159,7 +151,7 @@ QQuick3DViewport::QQuick3DViewport(QQuickItem *parent)
     m_sceneRoot = new QQuick3DSceneRootNode(this);
     m_environment = new QQuick3DSceneEnvironment(m_sceneRoot);
     m_renderStats = new QQuick3DRenderStats();
-    QQuick3DSceneManager *sceneManager = new QQuick3DSceneManager(m_sceneRoot);
+    QQuick3DSceneManager *sceneManager = new QQuick3DSceneManager();
     QQuick3DObjectPrivate::get(m_sceneRoot)->refSceneManager(*sceneManager);
     Q_ASSERT(sceneManager == QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager);
     connect(sceneManager, &QQuick3DSceneManager::needsUpdate,
@@ -174,17 +166,23 @@ QQuick3DViewport::QQuick3DViewport(QQuickItem *parent)
 
 QQuick3DViewport::~QQuick3DViewport()
 {
-    for (const auto &connection : qAsConst(m_connections))
+    // If the quick window still exists, make sure to disconnect any of the direct
+    // connections to this View3D
+    if (auto qw = window())
+        disconnect(qw, nullptr, this, nullptr);
+
+    for (const auto &connection : std::as_const(m_connections))
         disconnect(connection);
     auto sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
-    if (sceneManager)
+    if (sceneManager) {
         sceneManager->setParent(nullptr);
+        if (auto wa = sceneManager->wattached)
+            wa->queueForCleanup(sceneManager);
+    }
 
     delete m_sceneRoot;
     m_sceneRoot = nullptr;
 
-    delete sceneManager;
-
     // m_renderStats is tightly coupled with the render thread, so can't delete while we
     // might still be rendering.
     m_renderStats->deleteLater();
@@ -247,6 +245,9 @@ QQmlListProperty<QObject> QQuick3DViewport::data()
     This property specifies which \l Camera is used to render the scene. If this
     property is not set, then the first enabled camera in the scene will be used.
 
+    \note If this property contains a camera that's not \l {Node::visible}{visible} then
+    no further attempts to find a camera will be done.
+
     \sa PerspectiveCamera, OrthographicCamera, FrustumCamera, CustomCamera
 */
 QQuick3DCamera *QQuick3DViewport::camera() const
@@ -344,6 +345,32 @@ QQuick3DViewport::RenderMode QQuick3DViewport::renderMode() const
     return m_renderMode;
 }
 
+/*!
+    \qmlproperty enumeration QtQuick3D::View3D::renderFormat
+    \since 6.4
+
+    This property determines the backing texture's format. Applicable only when
+    the View3D is rendering to a texture, for example because the \l renderMode
+    is \c{View3D.Offscreen}.
+
+    The default is \c{ShaderEffectSource.RGBA8}.
+
+    If the format is not supported by the underlying graphics driver at run
+    time, RGBA8 is used.
+
+    \list
+    \li ShaderEffectSource.RGBA8
+    \li ShaderEffectSource.RGBA16F
+    \li ShaderEffectSource.RGBA32F
+    \endlist
+
+    \sa QtQuick::ShaderEffectSource::format, QtQuick::Item::layer.format
+ */
+QQuickShaderEffectSource::Format QQuick3DViewport::renderFormat() const
+{
+    return m_renderFormat;
+}
+
 /*!
     \qmlproperty QtQuick3D::RenderStats QtQuick3D::View3D::renderStats
     \readonly
@@ -386,11 +413,12 @@ QQuick3DSceneRenderer *QQuick3DViewport::createRenderer() const
                 // and this is the magic point where many things internally get
                 // switched over to be QRhi-based.
                 rhiContext->initialize(rhi);
-                rci = new QSSGRenderContextInterface(qw, rhiContext, QString::fromLatin1("./"));
+                rci = new QSSGRenderContextInterface(qw, rhiContext);
                 renderer = new QQuick3DSceneRenderer(rci);
             }
-
-            QObject::connect(qw, &QQuickWindow::afterFrameEnd, this, &QQuick3DViewport::cleanupResources);
+        } else {
+            qWarning("The Qt Quick scene is using a rendering method that is not based on QRhi and a 3D graphics API. "
+                     "Qt Quick 3D is not functional in such an environment. The View3D item is not going to display anything.");
         }
     }
 
@@ -501,6 +529,8 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
         if (!n->renderer) {
             n->window = window();
             n->renderer = createRenderer();
+            if (!n->renderer)
+                return nullptr;
             n->renderer->fboNode = n;
             n->quickFbo = this;
             connect(window(), SIGNAL(screenChanged(QScreen*)), n, SLOT(handleScreenChange()));
@@ -541,6 +571,8 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
         if (!n->renderer) {
             n->window = window();
             n->renderer = createRenderer();
+            if (!n->renderer)
+                return nullptr;
         }
 
         const QSize targetSize = window()->effectiveDevicePixelRatio() * QSize(width(), height());
@@ -549,7 +581,7 @@ QSGNode *QQuick3DViewport::updatePaintNode(QSGNode *node, QQuickItem::UpdatePain
         // { visible: false; layer.enabled: true } item still needs
         // to function normally.
         if (checkIsVisible() && isComponentComplete()) {
-            n->renderer->synchronize(this, targetSize, window()->effectiveDevicePixelRatio(), false);
+            n->renderer->synchronize(this, targetSize, window()->effectiveDevicePixelRatio());
             updateDynamicTextures();
             n->markDirty(QSGNode::DirtyMaterial);
         }
@@ -571,6 +603,8 @@ void QQuick3DViewport::itemChange(QQuickItem::ItemChange change, const QQuickIte
             if (m_importScene)
                 QQuick3DObjectPrivate::get(m_importScene)->sceneManager->setWindow(value.window);
         }
+    } else if (change == ItemVisibleHasChanged && isVisible()) {
+        update();
     }
 }
 
@@ -587,9 +621,14 @@ void QQuick3DViewport::setCamera(QQuick3DCamera *camera)
     if (m_camera == camera)
         return;
 
+    if (camera && !camera->parentItem())
+        camera->setParentItem(m_sceneRoot);
+    if (camera)
+        camera->updateGlobalVariables(QRect(0, 0, width(), height()));
+
+    QQuick3DObjectPrivate::attachWatcherPriv(m_sceneRoot, this, &QQuick3DViewport::setCamera, camera, m_camera);
+
     m_camera = camera;
-    if (m_camera)
-        m_camera->updateGlobalVariables(QRect(0, 0, width(), height()));
     emit cameraChanged();
     update();
 }
@@ -629,13 +668,18 @@ void QQuick3DViewport::setImportScene(QQuick3DNode *inScene)
 
     m_importScene = inScene;
     if (m_importScene) {
-        // If the referenced scene doesn't have a manager use the one from the sceen root (scenes defined outside of an view3d)
         auto privateObject = QQuick3DObjectPrivate::get(m_importScene);
         if (!privateObject->sceneManager) {
-            if (QQuick3DSceneManager *manager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager) {
+            // If object doesn't already have scene manager, check from its children
+            QQuick3DSceneManager *manager = findChildSceneManager(m_importScene);
+            // If still not found, use the one from the scene root (scenes defined outside of an view3d)
+            if (!manager)
+                manager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
+            if (manager) {
                 manager->setWindow(window());
                 privateObject->refSceneManager(*manager);
             }
+            // At this point some manager will exist
             Q_ASSERT(privateObject->sceneManager);
         }
 
@@ -670,6 +714,17 @@ void QQuick3DViewport::setRenderMode(QQuick3DViewport::RenderMode renderMode)
     update();
 }
 
+void QQuick3DViewport::setRenderFormat(QQuickShaderEffectSource::Format format)
+{
+    if (m_renderFormat == format)
+        return;
+
+    m_renderFormat = format;
+    m_renderModeDirty = true;
+    emit renderFormatChanged();
+    update();
+}
+
 
 /*!
     \qmlmethod vector3d View3D::mapFrom3DScene(vector3d scenePos)
@@ -743,43 +798,125 @@ QVector3D QQuick3DViewport::mapTo3DScene(const QVector3D &viewPos) const
     and return information about the nearest intersection with an object in the scene.
 
     This can, for instance, be called with mouse coordinates to find the object under the mouse cursor.
-
-    \note For an object with a custom \l Geometry, intersections will be approximated based on its bounding
-    box.
 */
 QQuick3DPickResult QQuick3DViewport::pick(float x, float y) const
 {
-    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio(),
-                           qreal(y) * window()->effectiveDevicePixelRatio());
-    // Some non-thread safe stuff to do input
-    // First need to get a handle to the renderer
-
     QQuick3DSceneRenderer *renderer = getRenderer();
     if (!renderer)
         return QQuick3DPickResult();
 
-    auto pickResult = renderer->syncPick(position);
-    if (!pickResult.m_hitObject)
+    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio(),
+                           qreal(y) * window()->effectiveDevicePixelRatio());
+    QSSGOption<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(position);
+    if (!rayResult.hasValue())
         return QQuick3DPickResult();
 
-    auto backendObject = pickResult.m_hitObject;
-    const auto sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
-    QQuick3DObject *frontendObject = sceneManager->lookUpNode(backendObject);
+    return processPickResult(renderer->syncPick(rayResult.getValue()));
 
-    // FIXME : for the case of consecutive importScenes
-    if (!frontendObject && m_importScene) {
-        const auto importSceneManager = QQuick3DObjectPrivate::get(m_importScene)->sceneManager;
-        frontendObject = importSceneManager->lookUpNode(backendObject);
-    }
+}
 
-    QQuick3DModel *model = qobject_cast<QQuick3DModel*>(frontendObject);
-    if (!model)
+/*!
+    \qmlmethod List<PickResult> View3D::pickAll(float x, float y)
+
+    This method will "shoot" a ray into the scene from view coordinates \a x and \a y
+    and return a list of information about intersections with objects in the scene.
+    The returned list is sorted by distance from the camera with the nearest
+    intersections appearing first and the furthest appearing last.
+
+    This can, for instance, be called with mouse coordinates to find the object under the mouse cursor.
+
+    \since 6.2
+*/
+QList<QQuick3DPickResult> QQuick3DViewport::pickAll(float x, float y) const
+{
+    QQuick3DSceneRenderer *renderer = getRenderer();
+    if (!renderer)
+        return QList<QQuick3DPickResult>();
+
+    const QPointF position(qreal(x) * window()->effectiveDevicePixelRatio(),
+                           qreal(y) * window()->effectiveDevicePixelRatio());
+    QSSGOption<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(position);
+    if (!rayResult.hasValue())
+        return QList<QQuick3DPickResult>();
+
+    const auto resultList = renderer->syncPickAll(rayResult.getValue());
+    QList<QQuick3DPickResult> processedResultList;
+    processedResultList.reserve(resultList.size());
+    for (const auto &result : resultList)
+        processedResultList.append(processPickResult(result));
+
+    return processedResultList;
+}
+
+/*!
+    \qmlmethod PickResult View3D::rayPick(vector3d origin, vector3d direction)
+
+    This method will "shoot" a ray into the scene starting at \a origin and in
+    \a direction and return information about the nearest intersection with an
+    object in the scene.
+
+    This can, for instance, be called with the position and forward vector of
+    any object in a scene to see what object is in front of an item. This
+    makes it possible to do picking from any point in the scene.
+
+    \since 6.2
+*/
+QQuick3DPickResult QQuick3DViewport::rayPick(const QVector3D &origin, const QVector3D &direction) const
+{
+    QQuick3DSceneRenderer *renderer = getRenderer();
+    if (!renderer)
         return QQuick3DPickResult();
 
-    return QQuick3DPickResult(model,
-                              ::sqrtf(pickResult.m_cameraDistanceSq),
-                              pickResult.m_localUVCoords,
-                              pickResult.m_scenePosition);
+    const QSSGRenderRay ray(origin, direction);
+
+    return processPickResult(renderer->syncPick(ray));
+}
+
+/*!
+    \qmlmethod List<PickResult> View3D::rayPickAll(vector3d origin, vector3d direction)
+
+    This method will "shoot" a ray into the scene starting at \a origin and in
+    \a direction and return a list of information about the nearest intersections with
+    objects in the scene.
+    The list is presorted by distance from the origin along the direction
+    vector with the nearest intersections appearing first and the furthest
+    appearing last.
+
+    This can, for instance, be called with the position and forward vector of
+    any object in a scene to see what objects are in front of an item. This
+    makes it possible to do picking from any point in the scene.
+
+    \since 6.2
+*/
+QList<QQuick3DPickResult> QQuick3DViewport::rayPickAll(const QVector3D &origin, const QVector3D &direction) const
+{
+    QQuick3DSceneRenderer *renderer = getRenderer();
+    if (!renderer)
+        return QList<QQuick3DPickResult>();
+
+    const QSSGRenderRay ray(origin, direction);
+
+    const auto resultList = renderer->syncPickAll(ray);
+    QList<QQuick3DPickResult> processedResultList;
+    processedResultList.reserve(resultList.size());
+    for (const auto &result : resultList)
+        processedResultList.append(processPickResult(result));
+
+    return processedResultList;
+}
+
+void QQuick3DViewport::processPointerEventFromRay(const QVector3D &origin, const QVector3D &direction, QPointerEvent *event)
+{
+    internalPick(event, origin, direction);
+}
+
+void QQuick3DViewport::setGlobalPickingEnabled(bool isEnabled)
+{
+    QQuick3DSceneRenderer *renderer = getRenderer();
+    if (!renderer)
+        return;
+
+    renderer->setGlobalPickingEnabled(isEnabled);
 }
 
 void QQuick3DViewport::invalidateSceneGraph()
@@ -787,25 +924,6 @@ void QQuick3DViewport::invalidateSceneGraph()
     m_node = nullptr;
 }
 
-void QQuick3DViewport::cleanupResources()
-{
-    // Pass the scene managers list of resouces marked for
-    // removal to the render context for deleation
-    // The render contect will take ownership of the nodes
-    // and clear the list
-    if (auto renderer = getRenderer()) {
-        const auto &rci = renderer->m_sgContext;
-        if (m_sceneRoot) {
-            const auto sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
-            rci->cleanupResources(sceneManager->resourceCleanupQueue);
-        }
-        if (m_importScene) {
-            const auto importSceneManager = QQuick3DObjectPrivate::get(m_importScene)->sceneManager;
-            rci->cleanupResources(importSceneManager->resourceCleanupQueue);
-        }
-    }
-}
-
 QQuick3DSceneRenderer *QQuick3DViewport::getRenderer() const
 {
     QQuick3DSceneRenderer *renderer = nullptr;
@@ -825,14 +943,14 @@ void QQuick3DViewport::updateDynamicTextures()
     // Must be called on the render thread.
 
     const auto &sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
-    for (auto *texture : qAsConst(sceneManager->qsgDynamicTextures))
+    for (auto *texture : std::as_const(sceneManager->qsgDynamicTextures))
         texture->updateTexture();
 
     QQuick3DNode *scene = m_importScene;
     while (scene) {
         const auto &importSm = QQuick3DObjectPrivate::get(scene)->sceneManager;
         if (importSm != sceneManager) {
-            for (auto *texture : qAsConst(importSm->qsgDynamicTextures))
+            for (auto *texture : std::as_const(importSm->qsgDynamicTextures))
                 texture->updateTexture();
         }
 
@@ -847,7 +965,10 @@ void QQuick3DViewport::setupDirectRenderer(RenderMode mode)
     auto renderMode = (mode == Underlay) ? QQuick3DSGDirectRenderer::Underlay
                                          : QQuick3DSGDirectRenderer::Overlay;
     if (!m_directRenderer) {
-        m_directRenderer = new QQuick3DSGDirectRenderer(createRenderer(), window(), renderMode);
+        QQuick3DSceneRenderer *sceneRenderer = createRenderer();
+        if (!sceneRenderer)
+            return;
+        m_directRenderer = new QQuick3DSGDirectRenderer(sceneRenderer, window(), renderMode);
         connect(window(), &QQuickWindow::sceneGraphInvalidated, this, &QQuick3DViewport::cleanupDirectRenderer, Qt::DirectConnection);
     }
 
@@ -855,7 +976,7 @@ void QQuick3DViewport::setupDirectRenderer(RenderMode mode)
     m_directRenderer->setViewport(QRectF(window()->effectiveDevicePixelRatio() * mapToScene(QPointF(0, 0)), targetSize));
     m_directRenderer->setVisibility(isVisible());
     if (isVisible()) {
-        m_directRenderer->renderer()->synchronize(this, targetSize.toSize(), window()->effectiveDevicePixelRatio(), false);
+        m_directRenderer->renderer()->synchronize(this, targetSize.toSize(), window()->effectiveDevicePixelRatio());
         updateDynamicTextures();
         m_directRenderer->requestRender();
     }
@@ -871,26 +992,39 @@ bool QQuick3DViewport::checkIsVisible() const
 
 }
 
-bool QQuick3DViewport::internalPick(QPointerEvent *event) const
+bool QQuick3DViewport::internalPick(QPointerEvent *event, const QVector3D &origin, const QVector3D &direction) const
 {
     // Some non-thread-safe stuff to do input
     // First need to get a handle to the renderer
     QQuick3DSceneRenderer *renderer = getRenderer();
-    if (!renderer)
+    if (!renderer || !event)
         return false;
 
-    qCDebug(lcEv) << event;
+    const bool isHover = QQuickDeliveryAgentPrivate::isHoverEvent(event);
+    if (!isHover)
+        qCDebug(lcEv) << event;
     struct SubsceneInfo {
         QQuick3DObject* obj = nullptr;
         QVarLengthArray<QPointF, 16> eventPointScenePositions;
     };
     QFlatMap<QQuickItem*, SubsceneInfo> visitedSubscenes;
 
+    const bool useRayPicking = !direction.isNull();
+
     for (int pointIndex = 0; pointIndex < event->pointCount(); ++pointIndex) {
-        auto eventPoint = QMutableEventPoint::from(event->point(pointIndex));
-        const QPointF realPosition = eventPoint.position() * window()->effectiveDevicePixelRatio();
-        QQuick3DSceneRenderer::PickResultList pickResults = renderer->syncPickAll(realPosition);
-        qCDebug(lcPick) << pickResults.count() << "pick results for" << event->point(pointIndex);
+        QQuick3DSceneRenderer::PickResultList pickResults;
+        auto &eventPoint = event->point(pointIndex);
+        if (useRayPicking) {
+            const QSSGRenderRay ray(origin, direction);
+            pickResults = renderer->syncPickAll(ray);
+        } else {
+            const QPointF realPosition = eventPoint.position() * window()->effectiveDevicePixelRatio();
+            QSSGOption<QSSGRenderRay> rayResult = renderer->getRayFromViewportPos(realPosition);
+            if (rayResult.hasValue())
+                pickResults = renderer->syncPickAll(rayResult.getValue());
+        }
+        if (!isHover)
+            qCDebug(lcPick) << pickResults.size() << "pick results for" << event->point(pointIndex);
         if (pickResults.isEmpty()) {
             eventPoint.setAccepted(false); // let it fall through the viewport to Items underneath
             continue; // next eventPoint
@@ -920,17 +1054,28 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
                 auto item2D = qobject_cast<QQuick3DItem2D *>(frontendObject);
                 if (item2D)
                     subsceneRootItem = item2D->contentItem();
+                if (!subsceneRootItem || subsceneRootItem->childItems().isEmpty())
+                    continue; // ignore empty 2D subscenes
                 // In this case the "UV" coordinates are in pixels in the subscene root item, so we can just use them.
                 subscenePosition = pickResult.m_localUVCoords.toPointF();
+                // Even though an Item2D is an "infinite plane" for rendering purposes,
+                // avoid delivering events outside the rectangular area that is occupied by child items,
+                // so that events can "fall through" to other interactive content in the scene or behind it.
+                if (!subsceneRootItem->childrenRect().contains(subscenePosition))
+                    continue;
             } else if (frontendObjectPrivate->type == QQuick3DObjectPrivate::Type::Model) {
                 // Model
                 int materialSubset = pickResult.m_subset;
                 const auto backendModel = static_cast<const QSSGRenderModel *>(backendObject);
                 // Get material
-                if (backendModel->materials.count() < (pickResult.m_subset + 1))
-                    materialSubset = backendModel->materials.count() - 1;
+                if (backendModel->materials.size() < (pickResult.m_subset + 1))
+                    materialSubset = backendModel->materials.size() - 1;
+                if (materialSubset < 0)
+                    continue;
                 const auto backendMaterial = backendModel->materials.at(materialSubset);
                 const auto frontendMaterial = sceneManager->lookUpNode(backendMaterial);
+                if (!frontendMaterial)
+                    continue;
                 const auto frontendMaterialPrivate = QQuick3DObjectPrivate::get(frontendMaterial);
 
                 if (frontendMaterialPrivate->type == QQuick3DObjectPrivate::Type::DefaultMaterial) {
@@ -950,15 +1095,26 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
                         if (principledMaterial->baseColorMap() && principledMaterial->baseColorMap()->sourceItem())
                             subsceneRootItem = principledMaterial->baseColorMap()->sourceItem();
                     }
+                } else if (frontendMaterialPrivate->type == QQuick3DObjectPrivate::Type::SpecularGlossyMaterial) {
+                    // SpecularGlossy Material
+                    const auto specularGlossyMaterial = qobject_cast<QQuick3DSpecularGlossyMaterial *>(frontendMaterial);
+                    if (specularGlossyMaterial) {
+                        // Just check for a albedoMap for now
+                        if (specularGlossyMaterial->albedoMap() && specularGlossyMaterial->albedoMap()->sourceItem())
+                            subsceneRootItem = specularGlossyMaterial->albedoMap()->sourceItem();
+                    }
                 } else if (frontendMaterialPrivate->type == QQuick3DObjectPrivate::Type::CustomMaterial) {
                     // Custom Material
                     const auto customMaterial = qobject_cast<QQuick3DCustomMaterial *>(frontendMaterial);
                     if (customMaterial) {
                         // This case is a bit harder because we can not know how the textures will be used
-                        for (const auto texture : customMaterial->dynamicTextureMaps()) {
-                            if (texture->sourceItem()) {
-                                subsceneRootItem = texture->sourceItem();
-                                break;
+                        const auto &texturesInputs = customMaterial->m_dynamicTextureMaps;
+                        for (const auto &textureInput : texturesInputs) {
+                            if (auto texture = textureInput->texture()) {
+                                if (texture->sourceItem()) {
+                                    subsceneRootItem = texture->sourceItem();
+                                    break;
+                                }
                             }
                         }
                     }
@@ -978,16 +1134,19 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
                 subscene.obj = frontendObject;
                 if (subscene.eventPointScenePositions.size() != event->pointCount()) {
                     // ensure capacity, and use an out-of-scene position rather than 0,0 by default
-                    // (if only QVarLengthArray.resize() would take an optional prototype argument...)
-                    subscene.eventPointScenePositions.resize(event->pointCount());
-                    for (auto &pt : subscene.eventPointScenePositions)
-                        pt = QPointF(-qInf(), -qInf());
+                    constexpr QPointF inf(-qt_inf(), -qt_inf());
+                    subscene.eventPointScenePositions.resize(event->pointCount(), inf);
                 }
                 subscene.eventPointScenePositions[pointIndex] = subscenePosition;
             }
-            qCDebug(lcPick) << "pick result:" << frontendObject << "@" << pickResult.m_scenePosition
-                            << "UV" << pickResult.m_localUVCoords << "dist" << qSqrt(pickResult.m_cameraDistanceSq)
-                            << "scene" << subscenePosition << subsceneRootItem;
+            if (isHover)
+                qCDebug(lcHover) << "hover pick result:" << frontendObject << "@" << pickResult.m_scenePosition
+                                << "UV" << pickResult.m_localUVCoords << "dist" << qSqrt(pickResult.m_distanceSq)
+                                << "scene" << subscenePosition << subsceneRootItem;
+            else
+                qCDebug(lcPick) << "pick result:" << frontendObject << "@" << pickResult.m_scenePosition
+                                << "UV" << pickResult.m_localUVCoords << "dist" << qSqrt(pickResult.m_distanceSq)
+                                << "scene" << subscenePosition << subsceneRootItem;
         } // for pick results from each QEventPoint
     } // for each QEventPoint
 
@@ -995,15 +1154,19 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
     // Maybe only some points fall inside, but QQuickDeliveryAgentPrivate::deliverMatchingPointsToItem()
     // makes reduced-subset touch events that contain only the relevant points, when necessary.
     bool ret = false;
-    auto mainDA = QQuickItemPrivate::get(window()->contentItem())->deliveryAgent();
 
 //    ViewportTransformHelper::removeAll();
+    QVarLengthArray<QPointF, 16> originalScenePositions;
+    originalScenePositions.resize(event->pointCount());
+    for (int pointIndex = 0; pointIndex < event->pointCount(); ++pointIndex)
+        originalScenePositions[pointIndex] = event->point(pointIndex).scenePosition();
     for (auto subscene : visitedSubscenes) {
         QQuickItem *subsceneRoot = subscene.first;
         auto &subsceneInfo = subscene.second;
-        Q_ASSERT(subsceneInfo.eventPointScenePositions.count() == event->pointCount());
+        Q_ASSERT(subsceneInfo.eventPointScenePositions.size() == event->pointCount());
         auto da = QQuickItemPrivate::get(subsceneRoot)->deliveryAgent();
-        qCDebug(lcPick) << "delivering to" << subsceneRoot << "via" << da << event;
+        if (!isHover)
+            qCDebug(lcPick) << "delivering to" << subsceneRoot << "via" << da << event;
 
         for (int pointIndex = 0; pointIndex < event->pointCount(); ++pointIndex) {
             const auto &pt = subsceneInfo.eventPointScenePositions.at(pointIndex);
@@ -1011,16 +1174,16 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
             // and Qt Quick expects it to arrive that way: then QQuickDeliveryAgentPrivate::translateTouchEvent()
             // copies it into the scene position before localizing.
             // That might be silly, we might change it eventually, but gotta stay consistent for now.
-            QMutableEventPoint &mut = QMutableEventPoint::from(event->point(pointIndex));
-            mut.setPosition(pt);
-            mut.setScenePosition(pt);
+            QEventPoint &ep = event->point(pointIndex);
+            QMutableEventPoint::setPosition(ep, pt);
+            QMutableEventPoint::setScenePosition(ep, pt);
         }
 
         if (event->isBeginEvent())
             da->setSceneTransform(nullptr);
         if (da->event(event)) {
             ret = true;
-            if (QQuickDeliveryAgentPrivate::anyPointGrabbed(event)) {
+            if (QQuickDeliveryAgentPrivate::anyPointGrabbed(event) && !useRayPicking) {
                 // In case any QEventPoint was grabbed, the relevant QQuickDeliveryAgent needs to know
                 // how to repeat the picking/coordinate transformation for each update,
                 // because delivery will bypass internalPick() due to the grab, and it's
@@ -1033,21 +1196,65 @@ bool QQuick3DViewport::internalPick(QPointerEvent *event) const
                 transform->targetItem = subsceneRoot;
                 transform->dpr = window()->effectiveDevicePixelRatio();
                 transform->uvCoordsArePixels = item2Dcase;
-                auto targetDA = (da->rootItem()->parentItem() ? mainDA : da);
-//                if (QQuickDeliveryAgentPrivate::isTouchEvent(event) && !item2Dcase)
-//                    targetDA = mainDA; // possible correction, but then TransformHelper has to map to different space?
-                transform->setOnDeliveryAgent(targetDA);
-                qCDebug(lcPick) << event->type() << "created ViewportTransformHelper on" << targetDA;
+                transform->setOnDeliveryAgent(da);
+                qCDebug(lcPick) << event->type() << "created ViewportTransformHelper on" << da;
             }
         } else if (event->type() != QEvent::HoverMove) {
             qCDebug(lcPick) << subsceneRoot << "didn't want" << event;
         }
         event->setAccepted(false); // reject implicit grab and let it keep propagating
     }
-
-    if (visitedSubscenes.isEmpty())
+    if (visitedSubscenes.isEmpty()) {
         event->setAccepted(false);
+    } else {
+        for (int pointIndex = 0; pointIndex < event->pointCount(); ++pointIndex)
+            QMutableEventPoint::setScenePosition(event->point(pointIndex), originalScenePositions.at(pointIndex));
+    }
     return ret;
 }
 
+QQuick3DPickResult QQuick3DViewport::processPickResult(const QSSGRenderPickResult &pickResult) const
+{
+    if (!pickResult.m_hitObject)
+        return QQuick3DPickResult();
+
+    auto backendObject = pickResult.m_hitObject;
+    const auto sceneManager = QQuick3DObjectPrivate::get(m_sceneRoot)->sceneManager;
+    QQuick3DObject *frontendObject = sceneManager->lookUpNode(backendObject);
+
+    // FIXME : for the case of consecutive importScenes
+    if (!frontendObject && m_importScene) {
+        const auto importSceneManager = QQuick3DObjectPrivate::get(m_importScene)->sceneManager;
+        frontendObject = importSceneManager->lookUpNode(backendObject);
+    }
+
+    QQuick3DModel *model = qobject_cast<QQuick3DModel *>(frontendObject);
+    if (!model)
+        return QQuick3DPickResult();
+
+    return QQuick3DPickResult(model,
+                              ::sqrtf(pickResult.m_distanceSq),
+                              pickResult.m_localUVCoords,
+                              pickResult.m_scenePosition,
+                              pickResult.m_localPosition,
+                              pickResult.m_faceNormal);
+}
+
+// Returns the first found scene manager of objects children
+QQuick3DSceneManager *QQuick3DViewport::findChildSceneManager(QQuick3DObject *inObject, QQuick3DSceneManager *manager)
+{
+    if (manager)
+        return manager;
+
+    auto children = QQuick3DObjectPrivate::get(inObject)->childItems;
+    for (auto child : children) {
+        if (auto m = QQuick3DObjectPrivate::get(child)->sceneManager) {
+            manager = m;
+            break;
+        }
+        manager = findChildSceneManager(child, manager);
+    }
+    return manager;
+}
+
 QT_END_NAMESPACE
